/*
 * dispatcher.c - short-term CPU scheduler for the SOS.
 */

#include "system.h"

/* Select and run a process. */
void Dispatcher(void) {

  /* Pick process to run. */
  current_process = SelectProcessToRun();

  /* Run it. */
  RunProcess(current_process);

  /* This function call will not return. */
}

/* Select process to run.  Returns -1 if none ready. */
int SelectProcessToRun() {
  /* Variable next_proc is static variable to "remember" last
     process selected and start searching from there. 
     Starts at max so wraps to first time called. */
  static int next_proc = MAX_PROCESSES;

  /* If process was interrupted by I/O interrupt or made 
     system call that did not block, then let it use rest of
     the time slice that was alloted to it. */
  if (current_process > 0 &&
      pd[current_process].slotAllocated &&
      pd[current_process].state == READY &&
      pd[current_process].timeLeft > 0)
    return current_process;

  /* Start from current value of next_proc and wrap at end. */
  for (int i=1; i < MAX_PROCESSES; i++) {
    next_proc++;
    if (next_proc >= MAX_PROCESSES)
      /* Note: slot 0, pid 0 reserved for OS, so wrap to 1. */
      next_proc = 1;

    /* next_proc is process we are considering. Check if is legit
       process and is ready to run. */
    if (pd[next_proc].slotAllocated && pd[next_proc].state == READY) {

      /* Found process. Give it an initial timeslice. */
      pd[next_proc].timeLeft = TIME_QUANTUM;

      /* Return it. */
      return next_proc;
    }
    
  }

  /* If here, then no process is ready to run. */
  return -1;
}

/* Run indicated process. */
void RunProcess(int pid) {

  /* Dispatch (run) the process (if one was picked). */
  if (pid > 0) {

    /* Set state to running. */
    pd[next_proc].state = RUNNING;

    /* Get time quantum available for process. */
    int quantum = pd[pid].timeLeft;

    /* Restore hardware state.  Includes assembly code since involves 
       hardware. */
    struct SaveArea *savearea = &(pd[pid].sa);
    asm {
      /* ia and psw taken from iia and ipsw by the rti instr. */
      load savearea+0, iia;
      load savearea+4, ipsw;
      /* Restore base and bound registers and saved values. */
      load savearea+8, base;
      load savearea+12, bound;
      /* Restore all the saved general registers. */
      loadall savearea+16;
      /* Set the interval timer */
      load quantum, timer;
      /* Return from interrupt restarts the process. */
      rti;
      /* rti does not return, since it sets the ia register. */
    }
    
  } else {
    /* If no process to run then must all be waiting for
       an interrupt. */
    while (1) /* no-op */;
    /* Note: control never drops out of the above loop. */
  }
}
